(* IR generation: translate takes a semantically checked AST and
   produces LLVM IR

   LLVM tutorial: Make sure to read the OCaml version of the tutorial

   http://llvm.org/docs/tutorial/index.html

   Detailed documentation on the OCaml LLVM library:

   http://llvm.moe/
   http://llvm.moe/ocaml/

*)

module L = Llvm
module A = Ast
open Sast

module StringMap = Map.Make(String)

(* translate : Sast.program -> Llvm.module *)
let translate program =
  let context    = L.global_context () in

  (* Create the LLVM compilation module into which
     we will generate code *)
  let the_module = L.create_module context "NanoC" in

  (* Get types from the context *)
  let i32_t      = L.i32_type    context
  and i1_t       = L.i1_type     context in

  (* Return the LLVM type for a NanoC type *)
  let ltype_of_typ = function
      A.Int   -> i32_t
    | A.Bool  -> i1_t
  in

  (* Create a map of global variables after creating each *)
  let global_vars : L.llvalue StringMap.t =
    let global_var m (t, n) =
      let init = L.const_int (ltype_of_typ t) 0
      in StringMap.add n (L.define_global n init the_module) m in
    List.fold_left global_var StringMap.empty program.slocals in

  (* Return the value for a variable or formal argument.
     Check global names *)
  let lookup n = StringMap.find n global_vars in

  (* Construct code for an expression; return its value *)
  let rec build_expr builder ((_, e) : sexpr) = match e with
      SLiteral i  -> L.const_int i32_t i
    | SBoolLit b  -> L.const_int i1_t (if b then 1 else 0)
    | SId s       -> L.build_load (lookup s) s builder
    | SAssign (s, e) -> let e' = build_expr builder e in
      ignore(L.build_store e' (lookup s) builder); e'
    | SBinop (e1, op, e2) ->
      let e1' = build_expr builder e1
      and e2' = build_expr builder e2 in
      (match op with
         A.Add     -> L.build_add
       | A.Sub     -> L.build_sub
       | A.And     -> L.build_and
       | A.Or      -> L.build_or
       | A.Equal   -> L.build_icmp L.Icmp.Eq
       | A.Neq     -> L.build_icmp L.Icmp.Ne
       | A.Less    -> L.build_icmp L.Icmp.Slt
      ) e1' e2' "tmp" builder
  in

  (* LLVM insists each basic block end with exactly one "terminator"
     instruction that transfers control.  This function runs "instr builder"
     if the current block does not already have a terminator.  Used,
     e.g., to handle the "fall off the end of the function" case. *)
  let add_terminal builder instr =
    match L.block_terminator (L.insertion_block builder) with
      Some _ -> ()
    | None -> ignore (instr builder) in

  (* Build the code for the given statement; return the builder for
     the statement's successor (i.e., the next instruction will be built
     after the one generated by this call) *)
  let rec build_stmt the_function builder = function
      SBlock sl -> List.fold_left (build_stmt the_function) builder sl
    | SExpr e -> ignore(build_expr builder e); builder
    | SIf (predicate, then_stmt, else_stmt) ->
      let bool_val = build_expr builder predicate in

      let then_bb = L.append_block context "then" the_function in
      ignore (build_stmt the_function (L.builder_at_end context then_bb) then_stmt);
      let else_bb = L.append_block context "else" the_function in
      ignore (build_stmt the_function (L.builder_at_end context else_bb) else_stmt);

      let end_bb = L.append_block context "if_end" the_function in
      let build_br_end = L.build_br end_bb in (* partial function *)
      add_terminal (L.builder_at_end context then_bb) build_br_end;
      add_terminal (L.builder_at_end context else_bb) build_br_end;

      ignore(L.build_cond_br bool_val then_bb else_bb builder);
      L.builder_at_end context end_bb

    | SWhile (predicate, body) ->
      let while_bb = L.append_block context "while" the_function in
      let build_br_while = L.build_br while_bb in (* partial function *)
      ignore (build_br_while builder);
      let while_builder = L.builder_at_end context while_bb in
      let bool_val = build_expr while_builder predicate in

      let body_bb = L.append_block context "while_body" the_function in
      add_terminal (build_stmt the_function (L.builder_at_end context body_bb) body) build_br_while;

      let end_bb = L.append_block context "while_end" the_function in

      ignore(L.build_cond_br bool_val body_bb end_bb while_builder);
      L.builder_at_end context end_bb

  in

  (* Fill in the body of the given function *)
  let build_program fdecl =
    (* Define each function (arguments and return type) so we can
       call it even before we've created its body *)
    let main_decl : L.llvalue =
      let ftype = L.function_type i32_t [||]
      in L.define_function "main" ftype the_module in
    let the_function = main_decl in

    let builder = L.builder_at_end context (L.entry_block the_function) in

    (* Build the code for each statement in the function *)
    let builder_end = List.fold_left (build_stmt the_function) builder fdecl.sbody in

    (* Add a return if the last block falls off the end *)
    add_terminal builder_end (L.build_ret (L.const_int i32_t 0))
  in

  build_program program;
  the_module
